//////////////////////////////////////////////////////////////////////
// ANSI																//
//////////////////////////////////////////////////////////////////////
// This unit interprets ANSI codes and turns them into cursor	   	//
// movements and color changes.  These are handled by virtual	  	//
// functions in a class derived from AnsiOutput.					//
//////////////////////////////////////////////////////////////////////

#include <cstring.h>
#include <unit/StrUtl.h>

#include <srcomm3/Ansi.h>

AnsiOutput::AnsiOutput()
{
	savex = savey = -1;
}

AnsiOutput::~AnsiOutput()
{
}

int AnsiOutput::put( char ch )
{
	// Pass all non-ANSI sequences straight through to screen
    if( ch != 27 && str.length() == 0 )
		return write( ch );

	// Interpret ESC character
	if( ch == 27 )
	{
		// Invalid ANSI sequences go straight through
        if( str.length() != 0 )
			write( str );

		str = "\033";
		return ch;
	}

	if( str == "\033" )
	{
		switch( ch )
		{
			case '[':
				// ANSI sequences must have a '['
				str += '[';
				break;
			default:
				// Invalid ANSI sequence
				write( str );
				write( ch );
				str = "";
		}
		return ch;
	}

	if( pos( ch, "0123456789;ABCDfHJKmsu" ) == -1 )
	{
		// Invalid ANSI sequence
		write( str );
		str = "";
		return write( ch );
	}

	if( pos( ch, "ABCDfHJKmsu" ) != -1 )
	{
		int numparam, params[100], maxparam = 100;
		parseparam( numparam, params, maxparam );

		switch( ch )
		{
			case 'A':
				if( numparam == 0 ) params[0] = 1;
				if( whereY() - params[0] < 1 )
					gotoxy( whereX(), 1 );
				else
					gotoxy( whereX(), whereY() - params[0] );
				break;
			case 'B':
				if( numparam == 0 ) params[0] = 1;
				if( whereX() + params[0] > height() )
					gotoxy( whereX(), height() );
				else
					gotoxy( whereX(), whereY() + params[0] );
				break;
			case 'C':
				if( numparam == 0 ) params[0] = 1;
				if( whereX() + params[0] > width() )
					gotoxy( width(), whereY() );
				else
					gotoxy( whereX() + params[0], whereY() );
				break;
			case 'D':
				if( numparam == 0 ) params[0] = 1;
				if( whereX() - params[0] < 1 )
					gotoxy( 1, whereY() );
				else   
					gotoxy( whereX() - params[0], whereY() );
				break;
			case 'H':
			case 'f':
			{
				int y = params[0];
				int x = params[1];
				if( numparam < 2 )   x = 1;
				if( numparam < 1 )   y = 1;
				if( x <= width() && x >= 1 && y <= height() && y >= 1 )
					gotoxy( x, y );
				break;
			}
			case 'J':
			{
				int p = params[0];
				if( numparam == 0 ) p = 0;

				if( p == 2 )	
					clrscr();
				if( p == 0 )
				{
					int oldx = whereX();
					int oldy = whereY();
					clrscr( oldy, height() );
					gotoxy( oldx, oldy );
				}
				if( p == 1 )
				{
					int oldx = whereX();
					int oldy = whereY();
					clrscr( 1, oldy );
					gotoxy( oldx, oldy );
				}
				break;
			}
			case 'K':
				clreol();
				break;
			case 'm':
			{
				int col = currentcolor();
				if( numparam == 0 )
					col = 0x07;
				for( int i = 0; i < numparam; i++ )
				{
					int p = params[i];
					switch( p )
					{
						case 0:
							col = 0x07;
							break;
						case 1:
							col |= 0x08;
							break;
						case 5:
							col |= 0x80;
							break;
						case 7:
							col = ( (col&15) << 4 ) | ( col >> 4 );
							break;
						case 30:
						case 31:
						case 32:
						case 33:
						case 34:
						case 35:
						case 36:
						case 37:
							p -= 30;
							col = ( col & 0xf8 ) |
								( ( p & 4 ) >> 2 ) |
								( p & 2 ) |
								( ( p & 1 ) << 2 );
							break;
						case 40:
						case 41:
						case 42:
						case 43:
						case 44:
						case 45:
						case 46:
						case 47:
							p -= 40;
							col = ( col & 0x8f ) |
								( ( p & 4 ) << 2 ) |
								( ( p & 2 ) << 4 ) |
								( ( p & 1 ) << 6 );
					}
				}
				setattr( col );
				break;
			}
			case 's':
				savey = whereY();
				savex = whereX();
				break;
			case 'u':
				if( savey > 0 && savex > 0 )
					gotoxy( savex, savey );
				break;
		}
		write( str );
		str = "";
		return ch;
	}

	if( ( ch == ';' ) || ( ch >= '0' && ch <= '9' ) )
		str += ch;

    if( str.length() > 256 )  // Must be an error!
	{
		write( str );
		str = "";
		return 0;
	}

	return ch;
}

void AnsiOutput::tabulate()
{
	int x = whereX();
	if( x < 80 )
	{
		do
			x++;
		while( x % 8 );
	}
	if( x == 80 )
		x = 1;
	gotoxy( x, whereY() );
	if( x == 1 )
		write( "\n\r" );
}

void AnsiOutput::backspace()
{
	if( whereX() > 1 )
		gotoxy( whereX() - 1, whereY() );
	else if( whereY() > 1 )
		gotoxy( width(), whereY() - 1 );
}

int AnsiOutput::write( char ch )
{
	switch( ch )
	{
		case 7:
			beep();
			break;
		case 8:
			backspace();
			break;
		case 9:
			tabulate();
			break;
		case 11:
			gotoxy( 1, 1 );
			break;
		case 12:
			clrscr();
			break;
		case 27:
			break;
		default:
			putchar( ch );
	 }
	 return ch;
}

void AnsiOutput::write( const string& s )
{
    for( int i = 0; i < s.length(); i++ )
		write( s[i] );
}

void AnsiOutput::parseparam( int& n, int values[], int len )
{
	n = 0;
	str.remove( 0, 2 );	   // ESC [ sequence
    while( str.length() > 0 && n < len )
	{
		string t;
        for( int i = 0; i < str.length() && str[i] != ';'; i++ )
			t += str[i];
		values[n++] = int( val(t) );
		str.remove( 0, i );
		if( str[0] == ';' ) str.remove( 0, 1 );
		if( i == 0 ) break;
	}
}
