// ASYNC.CPP unit
//
// Note:  This unit was translated from Pascal:
//     ASYNC2.PAS
//     Public domain async unit for doorgames by Joel Bergen
//     Gary Gordon & Joel Bergen added COM3 and COM4 support
//     Still needs a check for existence of a COM PORT in async_open()
//
// Modifications by Amit Patel:
//     I've added set_comport_parameters for COM1-COM8 support, and I have
//     added Desqview/MS-Windows time-slicing code to speed up operations
//     of other windows while we're doing busy-wait loops.
//
// Note from Amit Patel:
//     I've temporarily disabled the code to give up timeslices.

#include <srcomm3/async.h>

#include <dos.h>

static const unsigned char I8088_IMR = 0x21;
static unsigned baseport[8] = { 0x03f8, 0x02f8, 0x03e8, 0x02e8, 0, 0, 0, 0 };
static unsigned char asyncirq[8] = { 4, 3, 4, 3, 0, 0, 0, 0 };
static const int async_buffer_max = 1024;

static const int Ier = 1;
static const int Lcr = 3;
static const int Mcr = 4;
static const int Lsr = 5;
static const int Msr = 6;

static void interrupt (far *originalvector)(...); /* old interrupt function pointer */
static int
	originalLcr,
	originalImr,
	originalIer;

static char* async_buffer = 0;

static int async_open_flag = 0;
static int async_pause = 0;
static unsigned char async_port = 0;
static unsigned async_base;
static unsigned char async_irq;

static int async_buffer_overflow = 0;
static int async_buffer_used = 0;
static int async_buffer_head = 0;
static int async_buffer_tail = 0;
static int async_error_flag = 0;

static int async_checkcts = 1;

void interrupt __async_isr(...)
{
	enum { XON = 17, XOFF = 19 };

	char c;

// enable();
	if( async_buffer_used < async_buffer_max )
	{
		c = inportb( async_base );
		switch( c )
		{
			case XON:
				async_pause = 0;
				break;
			case XOFF:
				async_pause = 1;
				break;
			default:
				async_pause = 0;
				async_buffer[ async_buffer_head ] = c;
				if( async_buffer_head < async_buffer_max )
					async_buffer_head++;
				else
					async_buffer_head = 0;
				async_buffer_used++;
		}
	}
	else
		async_buffer_overflow = 1;
// disable();
	outportb( 0x20, 0x20 );
}

void Async::init()
{
	async_open_flag = 0;
	async_buffer_head = 0;
	async_buffer_tail = 0;
	async_buffer_overflow = 0;
	async_buffer_used = 0;
	async_pause = 0;
	async_checkcts = 1;
    async_buffer = new char[async_buffer_max+1];
}

void Async::close()
{
	if( async_open_flag )
	{
		disable();
		outportb( async_base + Ier, originalIer );
		outportb( async_base + Lcr, originalLcr );
		outportb( I8088_IMR, originalImr );
		enable();
		setvect( 8 + async_irq, originalvector );
		async_open_flag = 0;
	}
}

int Async::open( unsigned comport, long baudrate )
{
	int i, m;
	unsigned char b;

	if( async_open_flag )
		close();

	async_port = comport;
	async_base = baseport[ async_port-1 ];
	async_irq = asyncirq[ async_port-1 ];

	originalLcr = inportb( async_base + Lcr );

	outportb( async_base + Lcr, 0x03 );	// 0x03 == 8N1

	originalvector = getvect( 8 + async_irq );
	setvect( 8 + async_irq, __async_isr );

	disable();
	outportb( async_base + Lcr, inportb( async_base + Lcr ) & 0x7f );

	i = inportb( async_base + Lsr );
	i = inportb( async_base );
	i = inportb( I8088_IMR );

	originalImr = i;

	m = ( 1 << async_irq ) ^ 0x00ff;
	outportb( I8088_IMR, i & m );

	originalIer = inportb( async_base + Ier );
	outportb( async_base + Ier, 0x01 );
	i = inportb( async_base + Mcr );
	outportb( async_base + Mcr, i | 8 );
	enable();

	// Set baudrate
	b = inportb( async_base + Lcr ) | 128;
	outportb( async_base + Lcr, b );
	outportb( async_base, ( 115200L / baudrate ) & 0xff );
	outportb( async_base + 1, ( ( 115200L / baudrate ) >> 8 ) & 0xff );
	outportb( async_base + Lcr, b & 127 );

	async_open_flag = 1;

	return 1;
}

int Async::buffer_check()
{
	return ( async_buffer_used != 0 );
}

char Async::read()
{
	char c = async_buffer[ async_buffer_tail++ ];

	if( async_buffer_tail > async_buffer_max )
		async_buffer_tail = 0;
    disable();
	async_buffer_used--;
    enable();

	return c;
}

char Async::write( char c )
{
	unsigned t;

	if( !carrierdetect() )
		return 1;

	outportb( async_base + Mcr, 0x0b );

    if( async_checkcts )
    {
		for( t = 0; ( inportb( async_base + Msr ) & 0x10 ) == 0 && t < 10000; t++ )
			delay(1);	// wait for CTS

        if( t >= 10000 )
            return 1;
    }

    for( t = 0; ( inportb( async_base + Lsr ) & 0x20 ) == 0 && t < 10000; t++ )
		delay(1);
	if( t >= 10000 )
		return 1;

	t = 0;
	while( async_pause & carrierdetect() && t < 10000 )
	{
		// Wait for XON
		t++;
        delay( 10 );
	}

	if( t >= 10000 )
    {
        async_error_flag = 1;
        return 1;
    }

	if( async_pause )
		return 1;

	disable();
	outportb( async_base, c );
	enable();

	return 0;
}

void Async::hangup()
{
	outportb( async_base + Mcr, 0 );
    delay( 1000 );
	outportb( async_base + Mcr, 3 );
}

int Async::carrierdetect()
{
	int b = 1;
	unsigned w = 0;

	while( w < 500 && b )	{	// Make sure carrier stays down
		w++;			// and is not just a fluke
		b = ( inportb( async_base + Msr ) & 128 ) != 128;
	}
	return !b;
}

void __async_close()
{
    Async::close();
}
#pragma exit __async_close

void Async::set_comport_parameters( int port, int base, int irq )
{
    if( port < 1 || port > 8 )
        return;

    baseport[ port-1 ] = base;
    asyncirq[ port-1 ] = irq;
}
