//////////////////////////////////////////////////////////////////////
// FINOUT                                                           //
//////////////////////////////////////////////////////////////////////
// This unit provides text file input and output routines.  The     //
// fileout and filein classes are for write-only and read-only      //
// files.  There is no read-write text file, because I couldn't     //
// come up with good semantics.                                     //
//                                                                  //
// NOTE:  Unlike C++ Streams, the >> operator for strings will read //
//        an entire line, and it also filters out control chars.    //
//        The C++ Streams read a word at a time (which seems to be  //
//        less useful) and do not filter out control chars.         //
//////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <string.h>

#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <process.h>
#include <share.h>
#include <errno.h>
#include <dos.h>
#include <stdlib.h>

#include <srcomm3/inout.h>
#include <srcomm3/finout.h>

FileSystem::_ErrorHandler FileSystem::ErrorHandler = 0;

#define BUFFER_SIZE 512

///////////////////////////////////////////////////////////////////////////

fileout::fileout(): standardout( 0 )
{
    handle = -1;
	err = 0;
    binary = 0;
    buf_len = 0;
    buf_ptr = 0;
    buf = 0;
}

fileout::fileout( string filename, int appendflag, int binaryflag, int wide ): standardout( wide )
{
    handle = -1;
    err = 0;
    binary = binaryflag;
	buf_ptr = 0;
	buf_len = 0;
    buf = 0;

    getbuffer();

    open( filename, appendflag, binary );
}

fileout::~fileout()
{
	close();
	if( buf )       delete [] buf;
}

void fileout::getbuffer()
{
	if( buf )       return;

	buf_len = BUFFER_SIZE;
	buf = 0;
	while( buf_len > 16 && buf == 0 )
	{
		buf = new char[buf_len];
		if( !buf )      buf_len /= 2;
	}

	if( !buf )      buf_len = 0;
}

int fileout::open( string filename, int appendflag, int binaryflag )
{
	if( exists() )
		close();

	binary = binaryflag;

/******************************************************************************/
    binary = 1;         // SR inout does auto text->binary conversion
/******************************************************************************/

    handle = ::open( filename.c_str(), O_WRONLY | O_CREAT |
			( appendflag? O_APPEND : O_TRUNC ) |
			( binary? O_BINARY : O_TEXT ) |
			SH_DENYWR, S_IWRITE );

	if( handle < 0 )
	{
		handle = -1;
		err = errno;
		if( err == ENOENT )
			FileSystem::HandleError( "File '%s' not openable for output.\n", filename );
		if( err == EMFILE )
			FileSystem::HandleError( "Too many files open.  See manual.\n", "" );
//        if( err == EACCES )
//            ; // Probably a LOCK on the file

		return -1;
	}

	err = 0;

    // set the pointer to 0 so that the first write will
    // write into the beginning of the buffer
    buf_ptr = 0;

    getbuffer();

    return 0;
}

int fileout::close()
{
    if( !exists() )   return 0;

    flush();
    ::close( handle );

    handle = -1;
    err = 0;

    return 0;
}

int fileout::flush(void)
{
    if( !exists() )     return 0;

    int t;

    if( buf_ptr <= 0 || !buf )      return 0;

    t = buf_ptr;
	if( write( handle, buf, buf_ptr ) < 0 )
    {
        err = errno;
        t = -1;
    }

    buf_ptr = 0;

    return t;
}

void CC fileout::_put( int len, const char c[] )
{
    if( !exists() )  return;

//    getbuffer();

    if( buf_len == 0 )
    {
        if( write( handle, c, len ) == -1 )
            return;
    }
    else
    {
        for( int i = 0; i < len; i++ )
        {
            buf[buf_ptr] = c[i];
            buf_ptr++;
            if( buf_ptr >= buf_len )
                if( flush() == -1 )
                    return;
        }
    }
}

///////////////////////////////////////////////////////////////////////////

filein::filein()
{
    handle = -1;
	err = 0;

    binary = 0;
    buf_len = 0;
    buf_ptr = 0;
    buf = 0;
    buf_overflow = -1;
}

filein::filein( string filename, int binaryflag )
{
    handle = -1;
    err = 0;

    binary = binaryflag;
    buf_ptr = 0;
    buf_overflow = buf_len = 0;
    buf = 0;

    getbuffer();
	open( filename, binary );
}

filein::~filein()
{
	close();
	if( buf )       delete [] buf;
}

void filein::getbuffer()
{
    if( buf )       return;

    buf_len = BUFFER_SIZE;
    buf = 0;
    while( buf_len > 16 && buf == 0 )
    {
        buf = new char[buf_len];
        if( !buf )      buf_len /= 2;
    }
	buf_overflow = buf_len;
    buf_ptr = buf_len;

    if( !buf )      buf_len = 0;
}

int filein::open( string filename, int binaryflag )
{
    if( exists() )
        close();

    binary = binaryflag;

	handle = ::open( filename.c_str(), O_RDONLY |
            ( binary? O_BINARY : O_TEXT ) |
            SH_DENYNONE, S_IREAD );

    if( handle < 0 )
    {
        handle = -1;
		err = errno;

//		if( err == ENOENT )
//			; // file not found - no error message
		if( err == EMFILE )
			FileSystem::HandleError( "Too many files open.  Try increasing FILES= in CONFIG.SYS.\n", "" );
//		if( err == EACCES )
//			; // Probably a LOCK on the file

		return -1;
	}

	err = 0;

	// set the pointer to the length so that the first read will
    // read into the buffer
    getbuffer();

    buf_ptr = buf_len;
    buf_overflow = buf_len;

    return 0;
}

int filein::close()
{
	if( !exists() )   return 0;

	::close( handle );

	handle = -1;
	err = 0;

	return 0;
}

char CC filein::get()
{
    char c;

    if( !exists() )  return 0;

	if( buf_len == 0 )
	{
		int status = read( handle, &c, 1 );
		if( status == 1 )
			return c;
        err = errno;
        return 0;
    }

    if( buf_ptr >= buf_overflow )
    {
        int status = read( handle, buf, buf_len );

        if( status <= 0 )
        {
            status = 0;
            err = errno;
        }

        if( status < buf_len )
        {
            for( int i = status; i < buf_len; i++ )
                buf[i] = 0;
        }

        buf_overflow = status;
        buf_ptr = 0;
    }

    if( buf_overflow <= 0 )
    {
        err = 1;
        return 0;
    }

    c = buf[ buf_ptr ];
    buf_ptr++;

    return c;
}

char CC filein::out( char c )
{
	return c;
}

void CC filein::begininput()
{
	return;
}

void CC filein::endinput()
{
	return;
}

int CC filein::keyhit()
{
    if( !exists() )
        return 0;

    return ( err == 0 );
}

int filein::operator !()
{
    if( !exists() )  return 1;

    return ( err != 0 );
}

///////////////////////////////////////////////////////////////////////////

void FileSystem::HandleError( string msg, string arg )
{
    string str;
	for( int i = 0; msg[i]; i++ )
    {
		if( msg[i] == '%' && msg[i+1] == 's' )
		{
            str = str + arg;
            i++;
		}
		else
            str += msg[i];
    }
	if( ErrorHandler ) ErrorHandler( str );
}

void FileSystem::InstallErrorHandler( _ErrorHandler f )
{
	ErrorHandler = f;
}

FileSystem::_ErrorHandler FileSystem::GetErrorHandler()
{
	return ErrorHandler;
}
