// Game.h has some 'standard' things you might want to include in all
// your .cpp files
#include "game.h"

#include <string.h>
#include <stdlib.h>

// DoorGame is only for the "main" .CPP file, where you derive your own
// door game class from the standard door game class
#include <srcomm3/DoorGame.h>
// RecFile is for Turbo-Pascal-like record files
#include <srcomm3/RecFile.h>
// Door.h lets you find out information from the doorfile
#include <srcomm3/Door.h>

// This Game structure is just global information that you may want to
// keep around.  It's not part of SRCOMM -- I'm just using it for the
// sample program.  I'm using an array of chars instead of the string
// class because if you "dump" a struct to disk, it should not have any
// pointers in it.  The string class uses pointers, so the _pointer_
// would be saved, not the actual string.
struct Game
{
	char sysop[80];
};

Game GAME;

// To make a door game, you must derive from DoorGame and then override
// ::PlayGame().
class MyDoorGame: public DoorGame
{
  public:
    MyDoorGame(): DoorGame( "Castle Warfare", "CW", "1.10" ) {}

    virtual bool LocalParams();
	virtual void PlayGame();
};

int main()
{
	// Let's load up some data

	// RecordFile is a template class -- it is parameterized over a
	// class that you wish to store in a file
	RecordFile<Game> FGame( "DEMOGAME.DAT" );
	// If the file exists, and there is a record there
	if( FGame.filesize() >= 1 )
	{
		// We try to load record 0 from the file into GAME
		if( !FGame.load( GAME, 0 ) )
		{
			// If the return value is 0, it's an error.  For now,
			// let's assume it's a SHARE error
			kout << "It appears that another player is playing this game.\n";
			return 0;
		}
	}
	else
	{
		// Let's erase the GAME record
		GAME.sysop[0] = '\0';
	}

    // Let's start a new game and run it
    MyDoorGame().Run();

	// Save the data into the record file (position 0)
	FGame.save( GAME, 0 );

	return 0;
}

// This is one of the functions you must fill in.
void MyDoorGame::PlayGame()
{
	// Display a status bar
	statusBar -> Show();
	statusBar -> Display( DOOR, "CW", '*' );

	// If you use a lot of ANSI displays, you can uncomment this
	// and things will be more convenient.  However, having all output
	// go through the ANSI interpreter slows things down, so if you
	// only display ANSI in a few places, you can instead do what's
	// shown below (displaying the Title screen)
	//    if( GameOutput::Ansi )
	//        kout.setColorTranslator( AnsiTranslator );

	// Title screen
	if( DOOR.Ansi() )
	{
		// Turn off word wrap and pausing after 25 lines
		kout.SetRawMode();
		// Set up `kout' to understand ANSI codes
		bool UsingAnsi = kout.UsingAnsi();
		kout.SetAnsi( true );
		// Display the file
		displayfile( "TITLE.ANS" );
		// Restore the ANSI mode
		kout.SetAnsi( UsingAnsi );
		// Turn on word-wrap/pause mode
		kout.ResetRawMode();
	}
	else
		kout << "Welcome to Castle Warfare!\n\n";
	pause();

	// Do something silly, like run the game
	kout << "Hi there!!\n";
	kout << "Your sysop is " << GAME.sysop << "!\n";

	int soldiers = 5000;
	int computer = 5000;
	while( soldiers > 0 && computer > 0 )
	{
		// player's turn
		kout << attr(7) << "\n";
		{
			// I am creating a menu here
			// The numbers are the colors for the menu; see srmenu.h
			Menu M( "Menu1", "Castle Warfare", 15, 1, 9, 15, 11 );
			M.SetCols( 1 );		// one column
			// The first argument is a number that you want returned
			// when the user selects that menu option.  The second is
			// the key the user should press.  The third is the string;
			// if it is a blank string, the menu item is hidden from the
			// menu, but the option still works.  ('S' in this menu)
			// Note that more than one key can return the same number --
			// this lets you have `aliases' for commands.
			M.AddItem(  1, '1', "Attack guards" );
			M.AddItem(  2, '2', "Sneak past guards" );
			M.AddItem(  3, '3', "Status" );
			M.AddItem(  3, 'S', "" );
			M.AddItem(  9, 'Q', "Surrender" );

			int used_turn = 0;
			while( !used_turn )
			{
				// Menu::Go() returns -1 if the user pressed return
				int status = M.Go();
				switch( status )
				{
					case 1:
					{
						kout << "You attack Queen Quomputer's guards.\n";
						used_turn = 1;
						int computer_lost = 0;
						int soldiers_lost = 0;
						for( int i = 0; i < 100; i++ )
						{
							if( computer > 0 && random(100) < 90 )
							{
								computer--;
								computer_lost++;
							}
							if( soldiers > 0 && random(100) < 70 )
							{
								soldiers--;
								soldiers_lost++;
							}
						}
						kout << "You lost " << attr(13) << soldiers_lost
							 << attr(7) << " soldiers, and Queen Quomputer lost "
							 << attr(10) << computer_lost << attr(7) << " soldiers.\n";
						if( soldiers == 0 )
							kout << "All your soldiers are dead!\n";
						if( computer == 0 )
							kout << "All of Queen Quomputer's soldiers are dead!\n";
						break;
					}

					case 2:
					{
						kout << "You attempt to sneak past Queen Quomputer's guards...\n";
						used_turn = 1;
						for( int i = 0; i < computer; i++ )
						{
							if( random(100) == 0 )
							{
								kout << "You were caught!  Your soldier was lost.\n";
								soldiers--;
								// Eek!  A goto!
								goto failed1;
							}
						}
						kout << "The Queen is dead!  Long live the Queen!\n";
						computer = 0;
					  failed1:
						break;
					}

					case 3:
					{
						// do not set used_turn to 1
						kout << "You have " << attr(13) << soldiers << attr(7) << " soldiers.\n";
						kout << attr(3) << "Queen Quomputer has " << attr(10) << computer << attr(3) << " soldiers.\n";
						kout << attr(7) << '\n';
						break;
					}

					case 9:
					{
						kout << "You surrender to Queen Quomputer.\n";
						used_turn = 1;
						soldiers = 0;
						break;
					}
				}
			}
		}

		// Check if either team is dead
		if( computer <= 0 || soldiers <= 0 )
			break;

		kout << attr(3) << "\n";
		// Computer's turn
		switch( random(3) )
		{
			case 0:
			case 1:
			{
				kout << "Queen Quomputer attacks your guards!\n";

				int computer_lost = 0;
				int soldiers_lost = 0;
				for( int i = 0; i < 100; i++ )
				{
					if( computer > 0 && random(100) < 70 )
					{
						computer--;
						computer_lost++;
					}
					if( soldiers > 0 && random(100) < 90 )
					{
						soldiers--;
						soldiers_lost++;
					}
				}
				kout << "You lost " << attr(13) << soldiers_lost
					 << attr(3) << " soldiers, and Queen Quomputer lost "
					 << attr(10) << computer_lost << attr(3) << " soldiers.\n";
				if( soldiers == 0 )
					kout << "All your soldiers are dead!\n";
				if( computer == 0 )
					kout << "All of Queen Quomputer's soldiers are dead!\n";
				break;
			}
			case 2:
			{
				kout << "Queen Quomputer attempts to assasinate you!\n";
				for( int i = 0; i < soldiers; i++ )
				{
					if( random(100) == 0 )
					{
						kout << "The assasin was caught!\n";
						computer--;
						goto failed2;
					}
				}
				kout << "You have been assasinated!\n";
				soldiers = 0;
			  failed2:
				break;
			}
		}
	}

	if( soldiers <= 0 )
		kout << attr(14) << "\nYou have lost this game.\n";
	else
		kout << attr(11) << "\nYou have won.  Congratulations!\n";
}

// You can check for command line parameters here
// These are only LOCAL parameters, those which must be done
// without reading the doorfile.  If you want to check other
// parameters, you can do it up there in PlayGame().
bool MyDoorGame::LocalParams()
{
	if( Parameter( "Reset" ) )
	{
		string s;

		// Let's ask for the sysop name
		kout << "Sysop name:  ";
		kin >> s;
		strcpy( GAME.sysop, s.c_str() );

		kout << "The game has been reset!\n";

		// return 0 if you don't want to continue with the game
		return 0;
	}

	if( GAME.sysop[0] == '\0' )
	{
		kout << "Sorry, this game has not been reset.\n";
		// Notice that you can send different things to the local
		// and remote sides.  If there is nobody playing remotely,
		// the things sent remotely are ignored
		local << "Run   DEMOGAME RESET   to set up the game.\n";
		remote << "Please inform your sysop.\n";

        // return false because we don't want to continue
        return false;
	}

	// return true since we found no interesting parameters, so
	// we'd like to go on with the game
	return true;
}

// These two functions will go away soon; they are here for some hacky
// reasons.  They are used in SRE to implement `flavors'.
void initfilter()
{
}

int filterlength( const string& s )
{
	return length(s);
}
