// mtask.cpp
// Cooperative Multitasking in C++

#pragma hdrstop
#include <stdio.h>
#include <conio.h>
#include <time.h>
#include <dos.h>
#include <stdlib.h>

#include "mtask.h"

Task* Task::current = 0;
Task* Task::top = 0;
Task* Task::killed = 0;
int   Task::count = 0;

// Context switch flags
#define OK 1
#define KILLED 2

// Task.swap - switch context to the next task object in the round robin.
//	       It saves the current jump point with setjmp and then jumps
//	       to the next task's jump point with longjmp.
void Task::swap()
{
        if( !exists() )
                return;

	do
		current = current -> Next;
	while( current -> wakeup > clock() );

        volatile int status = setjmp( jmp );
        if( !status )
                longjmp( current -> jmp, OK );

        // The previous task was killed
        if( status == KILLED && killed )
	{
		delete killed;
                killed = 0;
        }
}

// _KillTask switches to the next task and passes a "KILLED" flag indicating
// that the current task should be removed
void _KillTask()
{
        if( Task::current -> Prev == Task::current )
        {
		// this is the only task left
                exit( 102 );
        }

        Task::killed = Task::current;
        Task::current = Task::current -> Prev;
	do
		Task::current = Task::current -> Prev;
	while( Task::current -> wakeup > clock()
	    || Task::current == Task::killed );

	longjmp( Task::current -> jmp, KILLED );
}

// Task.terminate - just calls _KillTask
void Task::terminate()
{
        _KillTask();
}

// Task constructor finds new stack space and puts the new function
// on the round robin.	If the function pointer is NULL, then the
// task is the main task, and no stack space is allocated
Task::Task( void (*f)(), int stacksize )
{
	// Keep track of how many tasks we have
        count++;

        // Set the task running
        wakeup = 0;

        volatile char *stacktop;

        if( f )
        {
                if( stacksize < 48 ) stacksize = 48;

		// Allocate new stack space
                stack = new char[stacksize];
                if( stack == 0 )
                {
                        printf( "\n\nError:  out of memory\n" );
                        exit( 101 );
                }

		// Put the return address of "Task::terminate" on the stack
		stacktop = stack + stacksize-8;

		// Figure out the memory model by the size of a pointer
		// Note this assumes LARGE or SMALL; it probably won't
		// work on the other memory models
#if sizeof(long*) == 2
			*((unsigned *)(stacktop)) = (unsigned)Task::terminate;
#else
			*((unsigned long *)(stacktop)) = (unsigned long)Task::terminate;
#endif
        }
        else
        {
                stack = 0;
                stacktop = 0;
        }

	// Save the jump point
        if( setjmp( jmp ) )
                return;

	// Set the stack switch registers in the jmp buffer
        if( f )
        {
                jmp -> j_sp = FP_OFF( stacktop );
                jmp -> j_ss = FP_SEG( stacktop );
                jmp -> j_bp = FP_OFF( stacktop );
                jmp -> j_cs = FP_SEG(f);
                jmp -> j_ip = FP_OFF(f);
        }
        else
                current = this;

	// Add this task onto the round robin list
	if( top )
        {
                Prev = top -> Prev;
                Next = top;
                top -> Prev -> Next = this;
                top -> Prev = this;
        }
        else
        {
                Next = this;
                Prev = this;
                top = this;
        }
}

// Task destructor removes itself from the round robin, and frees its
// stack space
Task::~Task()
{
        count--;

        Prev -> Next = Next;
        Next -> Prev = Prev;
        if( stack )
	{
                delete stack;
		stack = 0;
	}
        if( this == top )
                top = Next;
        Next = Prev = 0;
}

// ConTask.swap - saves the window, the colors, and the cursor location; then
//		  it swaps with Task.swap; when we return, we restore the
//		  window & color & cursor attributes
void ConTask::swap()
{
	text_info t;

        gettextinfo( &t );
        Task::swap();
        window( t.winleft, t.wintop, t.winright, t.winbottom );
        textattr( t.attribute );
        gotoxy( t.curx, t.cury );
}

// swap calls the current task's swap method
void swap()
{
        Task::current -> swap();
}

// wait is a delay for some number of milliseconds; it keeps calling
// swap() until it's time to wake up
void wait( long milliseconds )
{
	Task::current -> wakeup = clock() + milliseconds * 182L / 10000L;
        swap();
}

// mgetch runs other tasks until a key is available
int mgetch()
{
	while( !kbhit() )
		swap();
	return getch();
}

#ifdef DEMO

volatile int i = 0;

// SaveAndRestore saves a flag's old value, sets a new value, and restores
// it upon destruction; see f1() for an explanation
class SaveAndRestore
{
	int old;
	int *where;
public:
	SaveAndRestore( int& flag, int newval )
	{
		where = &flag;
		old = flag;
		flag = newval;
	}
	~SaveAndRestore()
	{
		*where = old;
	}
};

void showtime()
{
	static inf = 0;
	if( inf ) return;
        SaveAndRestore( inf, 1 );

	window( 1, 1, 80, 1 );
	textattr( 0x70 );

	for(;;)
	{
		char s[80];
		time_t t;
		time( &t );
		tm *tn = localtime( &t );

		gotoxy( 1, 1 );
		strftime( s, 80, " The Multitasking Demo:  It is %H:%M:%S %p on %A, %d %B %Y", tn );
                cprintf( "%s", s );
		clreol();

		wait( 1000 );
        }
}

void advertise()
{
	int msg = 0;
	static inf = 0;
	if( inf ) return;
	SaveAndRestore( inf, 1 );

	window( 1, 25, 80, 25 );
	textattr( 0x70 );

	for(;;)
	{
		gotoxy( 1, 1 );
		if( msg == 0 )
			cprintf( " Multitasking and Demo written by Amit Patel!!!" );
		if( msg == 1 )
			cprintf( " This Multitasking system was inspired by the work of Marc Tarpenning" );
		if( msg == 2 )
			cprintf( "                                Multitasking is fun" );
		clreol();

		wait( 3000 );
		msg = (msg+1) % 3;
	}
}

void f2()
{
	static int x = 1, y = 1;
        volatile int start = i;

	static inf = 0;
	if( inf ) return;
        SaveAndRestore( inf, 1 );

        // Set up the window & colors
	window( 41, 2, 80, 17 );
        textattr( 0x4f );
	if( y == 1 && x == 1 )
	{
                clrscr();
		cprintf( "This is window 2.  Press '2' to restart this task.\n\r" );
		x = wherex(); y = wherey();
	}

	// Run for 500 i's
	while( i < 500 + start )
	{
		gotoxy( x, y );
		cprintf( "B%d ", i++ );
		wait( 100 );			       // every 0.1 second
		x = wherex(); y = wherey();
	}
}

void f1()
{
	static int x = 1, y = 1;
        volatile int start = i;

	// We keep inf, "in function", as a flag so that we won't
	// run two of this function at the same time
	static inf = 0;
	if( inf ) return;
	// Save inf, set it to 1, and upon exit from this function (whereever
	// the function may exit), the destructor for SaveAndRestore is called,
	// so it restores inf back to 0
        SaveAndRestore( inf, 1 );

        window( 1, 2, 40, 17 );
        textattr( 0x1f );
	if( y == 1 && x == 1 )
	{
                clrscr();
		cprintf( "This is window 1.  Press '1' to restart this task.\n\r" );
		x = wherex(); y = wherey();
        }

	// Run for 60 i's
	while( i < 60 + start )
        {
                gotoxy( x, y );
                cprintf( "A%d ", i++ );
                swap();
		x = wherex(); y = wherey();
        }
}

int main()
{
        char c = 0;

	window( 1, 18, 80, 24 );
        textattr( 0x0e );
	clrscr();
	_setcursortype( _NOCURSOR );

	// Since our demo is based on screen windows & stuff, we'll
	// use ConTasks, which save the colors & windows & cursor locations
        new ConTask( 0 );
        new ConTask( f1 );
	new ConTask( f2 );
	new ConTask( showtime );
	new ConTask( advertise );

        while( c != 27 )
	{
		cprintf( "See the main routine run!   " );
                swap();                                // let other stuff run
                if( kbhit() )
                        c = getch();
                else
                        c = 0;

                if( c == '0' )
                        i = 0;
		if( c == '1' )
                        new ConTask( f1 );
		if( c == '2' )
			new ConTask( f2 );
        }

	window( 1, 1, 80, 25 );
	textattr( 0x07 );
	clrscr();
	_setcursortype( _NORMALCURSOR );

        return 0;
}

#endif

